from .app import AppConfig
from .cache import CacheConfig
from .compression import CompressionConfig
from .cors import CORSConfig
from .csrf import CSRFConfig
from .logging import BaseLoggingConfig, LoggingConfig, StructLoggingConfig
from .openapi import OpenAPIConfig
from .static_files import StaticFilesConfig
from .template import TemplateConfig

__all__ = (
    "AppConfig",
    "BaseLoggingConfig",
    "CORSConfig",
    "CSRFConfig",
    "CacheConfig",
    "CompressionConfig",
    "LoggingConfig",
    "OpenAPIConfig",
    "StaticFilesConfig",
    "StructLoggingConfig",
    "TemplateConfig",
)



import re
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, List, Optional, Pattern, Union

from pydantic import BaseConfig, BaseModel

from starlite.connection import ASGIConnection
from starlite.enums import ScopeType

if TYPE_CHECKING:
    from starlite.types import ASGIApp, Receive, Scope, Send


class AuthenticationResult(BaseModel):
    """This pydantic model is a container for authentication data."""

    user: Any
    """
    The user model, this can be any value corresponding to a user of the API
    """
    auth: Any = None
    """
    The auth value, this can for example be a JWT token.
    """

    class Config(BaseConfig):
        arbitrary_types_allowed = True


class AbstractAuthenticationMiddleware(ABC):
    scopes = {ScopeType.HTTP, ScopeType.WEBSOCKET}
    """
    Scopes supported by the middleware.
    """

    def __init__(
        self,
        app: "ASGIApp",
        exclude: Optional[Union[str, List[str]]] = None,
        exclude_from_auth_key: str = "exclude_from_auth",
    ) -> None:
        """This is an abstract AuthenticationMiddleware that allows users to
        create their own AuthenticationMiddleware by extending it and
        overriding the 'authenticate_request' method.

        Args:
            app: An ASGIApp, this value is the next ASGI handler to call in the middleware stack.
            exclude: A pattern or list of patterns to skip in the authentication middleware.
            exclude_from_auth_key: An identifier to use on routes to disable authentication for a particular route.
        """
        self.app = app
        self.exclude: Optional[Pattern[str]] = None
        self.exclude_from_auth_key = exclude_from_auth_key

        if exclude:
            self.exclude = re.compile("|".join(exclude)) if isinstance(exclude, list) else re.compile(exclude)

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """
        Args:
            scope: The ASGI connection scope.
            receive: The ASGI receive function.
            send: The ASGI send function.

        Returns:
            None
        """
        exclude_from_auth = scope["route_handler"].opt.get(self.exclude_from_auth_key)
        if (
            not exclude_from_auth
            and (not self.exclude or not self.exclude.findall(scope["path"]))
            and scope["type"] in self.scopes
        ):

            auth_result = await self.authenticate_request(ASGIConnection(scope))
            scope["user"] = auth_result.user
            scope["auth"] = auth_result.auth
        await self.app(scope, receive, send)

    @abstractmethod
    async def authenticate_request(self, connection: ASGIConnection) -> AuthenticationResult:  # pragma: no cover
        """This method must be overridden by subclasses. It receives the http
        connection and returns an instance of.

        Args:
            connection: An [ASGIConnection][starlite.connection.ASGIConnection] instance.

        Raises:
            NotAuthorizedException | PermissionDeniedException: if authentication fails.

        Returns:
            An instance of [AuthenticationResult][starlite.middleware.authentication.AuthenticationResult].
        """
        raise NotImplementedError("authenticate_request must be overridden by subclasses")



from typing import TYPE_CHECKING, Any, Callable, Dict, Type, Union

from typing_extensions import Literal

from starlite.types import Method

if TYPE_CHECKING:
    from starlite.app import Starlite  # noqa: TC004
    from starlite.controller import Controller  # noqa: TC004
    from starlite.handlers.asgi import ASGIRouteHandler  # noqa: TC004
    from starlite.handlers.http import HTTPRouteHandler  # noqa: TC004
    from starlite.handlers.websocket import WebsocketRouteHandler  # noqa: TC004
    from starlite.response import Response  # noqa: TC004
    from starlite.router import Router  # noqa: TC004
else:
    Starlite = Any
    ASGIRouteHandler = Any
    WebsocketRouteHandler = Any
    HTTPRouteHandler = Any
    Response = Any
    Controller = Any
    Router = Any

ReservedKwargs = Literal["request", "socket", "headers", "query", "cookies", "state", "data"]
StarliteType = Starlite
RouteHandlerType = Union[HTTPRouteHandler, WebsocketRouteHandler, ASGIRouteHandler]
ResponseType = Type[Response]
ControllerRouterHandler = Union[Type[Controller], RouteHandlerType, Router, Callable[..., Any]]
RouteHandlerMapItem = Union[WebsocketRouteHandler, ASGIRouteHandler, Dict[Method, HTTPRouteHandler]]



import re
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)
from uuid import UUID

from pydantic.datetime_parse import (
    parse_date,
    parse_datetime,
    parse_duration,
    parse_time,
)
from starlette.routing import Router as StarletteRouter

from starlite.enums import ScopeType
from starlite.exceptions import (
    MethodNotAllowedException,
    NotFoundException,
    ValidationException,
)
from starlite.utils import AsyncCallable

if TYPE_CHECKING:
    from starlite.app import HandlerNode, Starlite
    from starlite.routes.base import PathParameterDefinition
    from starlite.types import (
        ASGIApp,
        LifeSpanHandler,
        Receive,
        RouteHandlerType,
        Scope,
        Send,
    )


class PathParamNode:
    """Sentinel object to represent a path param in the route map."""


class PathParameterTypePathDesignator:
    """Sentinel object to a path parameter of type 'path'."""


PathParamPlaceholderType = Type[PathParamNode]
TerminusNodePlaceholderType = Type[PathParameterTypePathDesignator]
RouteMapNode = Dict[Union[str, PathParamPlaceholderType], Any]
ComponentsSet = Set[Union[str, PathParamPlaceholderType, TerminusNodePlaceholderType]]


class StarliteASGIRouter(StarletteRouter):
    """This class extends the Starlette Router class and *is* the ASGI app used
    in Starlite."""

    def __init__(
        self,
        app: "Starlite",
        on_shutdown: List["LifeSpanHandler"],
        on_startup: List["LifeSpanHandler"],
    ) -> None:
        self.app = app
        super().__init__(on_startup=on_startup, on_shutdown=on_shutdown)

    def _traverse_route_map(self, path: str, scope: "Scope") -> Tuple[RouteMapNode, List[str]]:
        """Traverses the application route mapping and retrieves the correct
        node for the request url.

        Args:
            path: The request's path.
            scope: The ASGI connection scope.

        Raises:
             NotFoundException: if no correlating node is found.

        Returns:
            A tuple containing the target RouteMapNode and a list containing all path parameter values.
        """
        path_params: List[str] = []
        current_node = self.app.route_map
        components = ["/", *[component for component in path.split("/") if component]]
        for idx, component in enumerate(components):
            components_set = cast("ComponentsSet", current_node["_components"])
            if component in components_set:
                current_node = cast("RouteMapNode", current_node[component])
                if "_static_path" in current_node:
                    self._handle_static_path(scope=scope, node=current_node)
                    break
                continue
            if PathParamNode in components_set:
                current_node = cast("RouteMapNode", current_node[PathParamNode])
                if PathParameterTypePathDesignator in components_set:
                    path_params.append("/".join(path.split("/")[idx:]))
                    break
                path_params.append(component)
                continue
            raise NotFoundException()
        return current_node, path_params

    @staticmethod
    def _handle_static_path(scope: "Scope", node: RouteMapNode) -> None:
        """Normalize the static path and update scope so file resolution will
        work as expected.

        Args:
            scope: The ASGI connection scope.
            node: Trie Node

        Returns:
            None
        """
        static_path = cast("str", node["_static_path"])
        if static_path != "/" and scope["path"].startswith(static_path):
            start_idx = len(static_path)
            scope["path"] = scope["path"][start_idx:] + "/"

    @staticmethod
    def _parse_path_parameters(
        path_parameter_definitions: List["PathParameterDefinition"], request_path_parameter_values: List[str]
    ) -> Dict[str, Any]:
        """Parses path parameters into their expected types.

        Args:
            path_parameter_definitions: A list of [PathParameterDefinition][starlite.route.base.PathParameterDefinition] instances
            request_path_parameter_values: A list of raw strings sent as path parameters as part of the request

        Raises:
            ValidationException: if path parameter parsing fails

        Returns:
            A dictionary mapping path parameter names to parsed values
        """
        result: Dict[str, Any] = {}
        parsers_map: Dict[Any, Callable] = {
            str: str,
            float: float,
            int: int,
            Decimal: Decimal,
            UUID: UUID,
            Path: lambda x: Path(re.sub("//+", "", (x.lstrip("/")))),
            date: parse_date,
            datetime: parse_datetime,
            time: parse_time,
            timedelta: parse_duration,
        }

        try:
            for idx, parameter_definition in enumerate(path_parameter_definitions):
                raw_param_value = request_path_parameter_values[idx]
                parameter_type = parameter_definition["type"]
                parameter_name = parameter_definition["name"]
                parser = parsers_map[parameter_type]
                result[parameter_name] = parser(raw_param_value)
            return result
        except (ValueError, TypeError, KeyError) as e:  # pragma: no cover
            raise ValidationException(
                f"unable to parse path parameters {','.join(request_path_parameter_values)}"
            ) from e

    def _parse_scope_to_route(self, scope: "Scope") -> Tuple[Dict[str, "HandlerNode"], bool]:
        """Given a scope object, retrieve the _asgi_handlers and _is_asgi
        values from correct trie node."""

        path = scope["path"].strip()
        if path != "/" and path.endswith("/"):
            path = path.rstrip("/")
        if path in self.app.plain_routes:
            current_node: RouteMapNode = self.app.route_map[path]
            path_params: List[str] = []
        else:
            current_node, path_params = self._traverse_route_map(path=path, scope=scope)

        scope["path_params"] = (
            self._parse_path_parameters(
                path_parameter_definitions=current_node["_path_parameters"], request_path_parameter_values=path_params
            )
            if path_params
            else {}
        )

        asgi_handlers = cast("Dict[str, HandlerNode]", current_node["_asgi_handlers"])
        is_asgi = cast("bool", current_node["_is_asgi"])
        return asgi_handlers, is_asgi

    @staticmethod
    def _resolve_handler_node(
        scope: "Scope", asgi_handlers: Dict[str, "HandlerNode"], is_asgi: bool
    ) -> Tuple["ASGIApp", "RouteHandlerType"]:
        """Given a scope, returns the ASGI App and route handler for the
        route."""
        if is_asgi:
            node = asgi_handlers[ScopeType.ASGI]
        elif scope["type"] == ScopeType.HTTP:
            if scope["method"] not in asgi_handlers:
                raise MethodNotAllowedException()
            node = asgi_handlers[scope["method"]]
        else:
            node = asgi_handlers[ScopeType.WEBSOCKET]
        return node["asgi_app"], node["handler"]

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:  # type: ignore[override]
        """The main entry point to the Router class."""
        try:
            asgi_handlers, is_asgi = self._parse_scope_to_route(scope=scope)
            asgi_app, handler = self._resolve_handler_node(scope=scope, asgi_handlers=asgi_handlers, is_asgi=is_asgi)
        except KeyError as e:
            raise NotFoundException() from e
        scope["route_handler"] = handler
        await asgi_app(scope, receive, send)

    async def _call_lifespan_handler(self, handler: "LifeSpanHandler") -> None:
        """Determines whether the lifecycle handler expects an argument, and if
        so passes the `app.state` to it. If the handler is an async function,
        it awaits the return.

        Args:
            handler (LifeSpanHandler): sync or async callable that may or may not have an argument.
        """
        async_callable = AsyncCallable(handler)  # type: ignore

        if async_callable.num_expected_args > 0:
            await async_callable(self.app.state)  # type: ignore[arg-type]
        else:
            await async_callable()

    async def startup(self) -> None:
        """Run any [LifeSpanHandlers][starlite.types.LifeSpanHandler] defined
        in the application's `.on_startup` list.

        Calls the `before_startup` hook and `after_startup` hook
        handlers respectively before and after calling in the lifespan
        handlers.
        """
        for hook in self.app.before_startup:
            await hook(self.app)

        for handler in self.on_startup:
            await self._call_lifespan_handler(handler)

        for hook in self.app.after_startup:
            await hook(self.app)

    async def shutdown(self) -> None:
        """Run any [LifeSpanHandlers][starlite.types.LifeSpanHandler] defined
        in the application's `.on_shutdown` list.

        Calls the `before_shutdown` hook and `after_shutdown` hook
        handlers respectively before and after calling in the lifespan
        handlers.
        """

        for hook in self.app.before_shutdown:
            await hook(self.app)

        for handler in self.on_shutdown:
            await self._call_lifespan_handler(handler)

        for hook in self.app.after_shutdown:
            await hook(self.app)



from typing import Any

from starlite.exceptions.base_exceptions import StarLiteException


class WebSocketException(StarLiteException):
    code: int
    """Exception code. Should be a number in the 4000+ range."""

    def __init__(self, *args: Any, detail: str, code: int = 4500) -> None:
        """Exception class for websocket related events.

        Args:
            *args: Any exception args.
            detail:
            code: Exception code. Should be a number in the 4000+ range.
        """
        super().__init__(*args, detail=detail)
        self.code = code



# pylint: disable=unused-argument, import-outside-toplevel
import os
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    AsyncIterable,
    AsyncIterator,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

from pydantic import BaseConfig, FilePath, validator
from pydantic.generics import GenericModel
from starlette.responses import FileResponse, RedirectResponse
from starlette.responses import Response as StarletteResponse
from starlette.responses import StreamingResponse

from starlite.datastructures.background_tasks import BackgroundTask, BackgroundTasks
from starlite.datastructures.cookie import Cookie
from starlite.enums import MediaType

if TYPE_CHECKING:

    from starlite.app import Starlite
    from starlite.connection import Request
    from starlite.response import TemplateResponse


R = TypeVar("R", bound=StarletteResponse)


class ResponseContainer(GenericModel, ABC, Generic[R]):
    class Config(BaseConfig):
        arbitrary_types_allowed = True

    background: Optional[Union[BackgroundTask, BackgroundTasks]] = None
    """
        A [BackgroundTask][starlite.datastructures.BackgroundTask] instance or
        [BackgroundTasks][starlite.datastructures.BackgroundTasks] to execute after the response is finished.
        Defaults to None.
    """
    headers: Dict[str, Any] = {}
    """A string/string dictionary of response headers. Header keys are insensitive. Defaults to None."""
    cookies: List[Cookie] = []
    """A list of Cookie instances to be set under the response 'Set-Cookie' header. Defaults to None."""
    media_type: Optional[Union[MediaType, str]] = None
    """If defined, overrides the media type configured in the route decorator"""

    @abstractmethod
    def to_response(
        self,
        headers: Dict[str, Any],
        media_type: Union["MediaType", str],
        status_code: int,
        app: "Starlite",
        request: "Request",
    ) -> R:  # pragma: no cover
        """Abstract method that should be implemented by subclasses. Returns a
        Starlette compatible Response instance.

        Args:
            headers: A dictionary of headers.
            media_type: A string or member of the [MediaType][starlite.enums.MediaType] enum.
            status_code: A response status code.
            app: The [Starlite][starlite.app.Starlite] application instance.
            request: A [Request][starlite.connection.request.Request] instance.

        Returns:
            A Response Object
        """
        raise NotImplementedError("not implemented")


class File(ResponseContainer[FileResponse]):
    """Container type for returning File responses."""

    path: FilePath
    """Path to the file to send"""
    filename: str
    """The filename"""
    stat_result: Optional[os.stat_result] = None
    """File statistics"""

    @validator("stat_result", always=True)
    def validate_status_code(  # pylint: disable=no-self-argument
        cls, value: Optional[os.stat_result], values: Dict[str, Any]
    ) -> os.stat_result:
        """Set the stat_result value for the given filepath."""
        return value or os.stat(cast("str", values.get("path")))

    def to_response(
        self,
        headers: Dict[str, Any],
        media_type: Union["MediaType", str],
        status_code: int,
        app: "Starlite",
        request: "Request",
    ) -> FileResponse:
        """Creates a FileResponse instance.

        Args:
            headers: A dictionary of headers.
            media_type: A string or member of the [MediaType][starlite.enums.MediaType] enum.
            status_code: A response status code.
            app: The [Starlite][starlite.app.Starlite] application instance.
            request: A [Request][starlite.connection.request.Request] instance.

        Returns:
            A FileResponse instance
        """
        return FileResponse(
            background=self.background,
            filename=self.filename,
            headers=headers,
            media_type=media_type,
            path=self.path,
            stat_result=self.stat_result,
            status_code=status_code,
        )


class Redirect(ResponseContainer[RedirectResponse]):
    """Container type for returning Redirect responses."""

    path: str
    """Redirection path"""

    def to_response(
        self,
        headers: Dict[str, Any],
        media_type: Union["MediaType", str],
        status_code: int,
        app: "Starlite",
        request: "Request",
    ) -> RedirectResponse:
        """Creates a RedirectResponse instance.

        Args:
            headers: A dictionary of headers.
            media_type: A string or member of the [MediaType][starlite.enums.MediaType] enum.
            status_code: A response status code.
            app: The [Starlite][starlite.app.Starlite] application instance.
            request: A [Request][starlite.connection.request.Request] instance.

        Returns:
            A RedirectResponse instance
        """
        return RedirectResponse(headers=headers, status_code=status_code, url=self.path, background=self.background)


class Stream(ResponseContainer[StreamingResponse]):
    """Container type for returning Stream responses."""

    iterator: Union[
        Iterator[Union[str, bytes]],
        Generator[Union[str, bytes], Any, Any],
        AsyncIterator[Union[str, bytes]],
        AsyncGenerator[Union[str, bytes], Any],
        Type[Iterator[Union[str, bytes]]],
        Type[AsyncIterator[Union[str, bytes]]],
        Callable[[], AsyncGenerator[Union[str, bytes], Any]],
        Callable[[], Generator[Union[str, bytes], Any, Any]],
    ]
    """Iterator, Generator or async Iterator or Generator returning stream chunks"""

    def to_response(
        self,
        headers: Dict[str, Any],
        media_type: Union["MediaType", str],
        status_code: int,
        app: "Starlite",
        request: "Request",
    ) -> StreamingResponse:
        """Creates a StreamingResponse instance.

        Args:
            headers: A dictionary of headers.
            media_type: A string or member of the [MediaType][starlite.enums.MediaType] enum.
            status_code: A response status code.
            app: The [Starlite][starlite.app.Starlite] application instance.
            request: A [Request][starlite.connection.request.Request] instance.

        Returns:
            A StreamingResponse instance
        """

        return StreamingResponse(
            background=self.background,
            content=self.iterator if isinstance(self.iterator, (Iterable, AsyncIterable)) else self.iterator(),
            headers=headers,
            media_type=media_type,
            status_code=status_code,
        )


class Template(ResponseContainer["TemplateResponse"]):
    """Container type for returning Template responses."""

    name: str
    """Path-like name for the template to be rendered, e.g. "index.html"."""
    context: Optional[Dict[str, Any]] = None
    """A dictionary of key/value pairs to be passed to the temple engine's render method. Defaults to None."""

    def to_response(
        self,
        headers: Dict[str, Any],
        media_type: Union["MediaType", str],
        status_code: int,
        app: "Starlite",
        request: "Request",
    ) -> "TemplateResponse":
        """Creates a TemplateResponse instance.

        Args:
            headers: A dictionary of headers.
            media_type: A string or member of the [MediaType][starlite.enums.MediaType] enum.
            status_code: A response status code.
            app: The [Starlite][starlite.app.Starlite] application instance.
            request: A [Request][starlite.connection.request.Request] instance

        Raises:
            [ImproperlyConfiguredException][starlite.exceptions.ImproperlyConfiguredException]: if app.template_engine
                is not configured.

        Returns:
            A TemplateResponse instance
        """
        from starlite.exceptions import ImproperlyConfiguredException
        from starlite.response import TemplateResponse

        context = self.context or {}
        if not app.template_engine:
            raise ImproperlyConfiguredException("Template engine is not configured")
        return TemplateResponse(
            background=self.background,
            context={**context, "request": request},
            headers=headers,
            status_code=status_code,
            template_engine=app.template_engine,
            template_name=self.name,
        )



from functools import partial
from inspect import getfullargspec, ismethod
from typing import Any, Callable, Dict, Generic, List, TypeVar, Union, cast

from anyio.to_thread import run_sync
from typing_extensions import Literal, ParamSpec

from starlite.utils.predicates import is_async_callable

P = ParamSpec("P")
T = TypeVar("T")


class AsyncCallable(Generic[P, T]):
    __slots__ = ("args", "kwargs", "wrapped_callable", "is_method", "num_expected_args")

    def __init__(self, fn: Callable[P, T]) -> None:
        """Utility class that wraps a callable and ensures it can be called as
        an async function.

        Args:
            fn: Callable to wrap - can be any sync or async callable.
        """

        self.is_method = ismethod(fn)
        self.num_expected_args = len(getfullargspec(fn).args) - (1 if self.is_method else 0)
        self.wrapped_callable: Dict[Literal["fn"], Callable] = {
            "fn": fn if is_async_callable(fn) else async_partial(fn)
        }

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        """A proxy to the wrapped function's call method.

        Args:
            *args: Args of the wrapped function.
            **kwargs: Kwargs of the wrapper function.

        Returns:
            The return value of the wrapped function.
        """
        return cast("T", await self.wrapped_callable["fn"](*args, **kwargs))


def as_async_callable_list(value: Union[Callable, List[Callable]]) -> List[AsyncCallable]:
    """
    Helper function to handle wrapping values in AsyncCallables
    Args:
        value: A callable or list of callables.

    Returns:
        A list of AsyncCallable instances
    """
    if not isinstance(value, list):
        return [AsyncCallable(value)]
    return [AsyncCallable(v) for v in value]


def async_partial(fn: Callable) -> Callable:
    """This function wraps a given sync function making it async. In difference
    to the 'asyncio.run_sync' function, it allows for passing kwargs.

    Args:
        fn: A sync callable to wrap.

    Returns:
        A wrapper
    """

    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        applied_kwarg = partial(fn, **kwargs)
        return await run_sync(applied_kwarg, *args)

    return wrapper



import binascii
import contextlib
import time
from base64 import b64decode, b64encode
from os import urandom
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, List, Optional, cast

from orjson import OPT_SERIALIZE_NUMPY, dumps, loads
from pydantic import (
    BaseConfig,
    BaseModel,
    SecretBytes,
    conint,
    conlist,
    constr,
    validator,
)
from starlette.datastructures import MutableHeaders
from typing_extensions import Literal

from starlite.connection import ASGIConnection
from starlite.datastructures.cookie import Cookie
from starlite.exceptions import MissingDependencyException
from starlite.middleware.base import DefineMiddleware, MiddlewareProtocol
from starlite.utils import get_serializer_from_scope
from starlite.utils.serialization import default_serializer

try:
    from cryptography.exceptions import InvalidTag
    from cryptography.hazmat.primitives.ciphers.aead import AESGCM
except ImportError as e:
    raise MissingDependencyException("cryptography is not installed") from e

if TYPE_CHECKING:
    from starlite.types import ASGIApp, Message, Receive, Scope, Send


ONE_DAY_IN_SECONDS = 60 * 60 * 24
NONCE_SIZE = 12
CHUNK_SIZE = 4096 - 64
AAD = b"additional_authenticated_data="


class SessionCookieConfig(BaseModel):
    """Configuration for Session middleware."""

    class Config(BaseConfig):
        arbitrary_types_allowed = True

    secret: SecretBytes
    """
    A secret key to use for generating an encryption key.
    Must have a length of 16 (128 bits), 24 (192 bits) or 32 (256 bits) characters.
    """
    key: constr(min_length=1, max_length=256) = "session"  # type: ignore[valid-type]
    """
    Key to use for the cookie inside the header,
    e.g. `session=<data>` where 'session' is the cookie key and <data> is the session data.

    Notes:
        - If a session cookie exceeds 4KB in size it is split. In this case the key will be of the format
            'session-{segment number}'.
    """
    max_age: conint(ge=1) = ONE_DAY_IN_SECONDS * 14  # type: ignore[valid-type]
    """Maximal age of the cookie before its invalidated."""
    scopes: conlist(Literal["http", "websocket"], min_items=1, max_items=2) = ["http", "websocket"]  # type: ignore[valid-type]
    """Scopes for the middleware - options are 'http' and 'websocket' with the default being both"""
    path: str = "/"
    """Path fragment that must exist in the request url for the cookie to be valid. Defaults to '/'."""
    domain: Optional[str] = None
    """Domain for which the cookie is valid."""
    secure: bool = False
    """Https is required for the cookie."""
    httponly: bool = True
    """Forbids javascript to access the cookie via 'Document.cookie'."""
    samesite: Literal["lax", "strict", "none"] = "lax"
    """Controls whether or not a cookie is sent with cross-site requests. Defaults to 'lax'."""

    @validator("secret", always=True)
    def validate_secret(cls, value: SecretBytes) -> SecretBytes:  # pylint: disable=no-self-argument
        """Ensures that the 'secret' value is 128, 192 or 256 bits.

        Args:
            value: A bytes string.

        Raises:
            ValueError: if the bytes string is of incorrect length.

        Returns:
            A bytes string.
        """
        if len(value.get_secret_value()) not in {16, 24, 32}:
            raise ValueError("secret length must be 16 (128 bit), 24 (192 bit) or 32 (256 bit)")
        return value

    @property
    def middleware(self) -> DefineMiddleware:
        """Use this property to insert the config into a middleware list on one
        of the application layers.

        Examples:

            ```python
            from os import urandom

            from starlite import Starlite, Request, get
            from starlite.middleware.session import SessionCookieConfig

            session_config = SessionCookieConfig(secret=urandom(16))


            @get("/")
            def my_handler(request: Request) -> None:
                ...


            app = Starlite(route_handlers=[my_handler], middleware=[session_config.middleware])
            ```

        Returns:
            An instance of DefineMiddleware including 'self' as the config kwarg value.
        """
        return DefineMiddleware(SessionMiddleware, config=self)


class SessionMiddleware(MiddlewareProtocol):
    def __init__(
        self,
        app: "ASGIApp",
        config: SessionCookieConfig,
    ) -> None:
        """Starlite SessionMiddleware.

        Args:
            app: The 'next' ASGI app to call.
            config: SessionCookieConfig instance.
        """
        self.app = app
        self.config = config
        self.aesgcm = AESGCM(config.secret.get_secret_value())

    def dump_data(self, data: Any, scope: Optional["Scope"] = None) -> List[bytes]:
        """Given orjson serializable data, including pydantic models and numpy
        types, dump it into a bytes string, encrypt, encode and split it into
        chunks of the desirable size.

        Args:
            data: Data to serialize, encrypt, encode and chunk.
            scope: The ASGI connection scope.

        Notes:
            - The returned list is composed of a chunks of a single base64 encoded
                string that is encrypted using AES-CGM.

        Returns:
            List of encoded bytes string of a maximum length equal to the 'CHUNK_SIZE' constant.
        """
        serializer = (get_serializer_from_scope(scope) if scope else None) or default_serializer
        serialized = dumps(data, default=serializer, option=OPT_SERIALIZE_NUMPY)
        associated_data = dumps({"expires_at": round(time.time()) + self.config.max_age})
        nonce = urandom(NONCE_SIZE)
        encrypted = self.aesgcm.encrypt(nonce, serialized, associated_data=associated_data)
        encoded = b64encode(nonce + encrypted + AAD + associated_data)
        return [encoded[i : i + CHUNK_SIZE] for i in range(0, len(encoded), CHUNK_SIZE)]

    def load_data(self, data: List[bytes]) -> Dict[str, Any]:
        """Given a list of strings, decodes them into the session object.

        Args:
            data: A list of strings derived from the request's session cookie(s).

        Returns:
            A deserialized session value.
        """
        decoded = b64decode(b"".join(data))
        nonce = decoded[:NONCE_SIZE]
        aad_starts_from = decoded.find(AAD)
        associated_data = decoded[aad_starts_from:].replace(AAD, b"") if aad_starts_from != -1 else None
        if associated_data and loads(associated_data)["expires_at"] > round(time.time()):
            encrypted_session = decoded[NONCE_SIZE:aad_starts_from]
            decrypted = self.aesgcm.decrypt(nonce, encrypted_session, associated_data=associated_data)
            return cast("Dict[str, Any]", loads(decrypted))
        return {}

    def create_send_wrapper(
        self, scope: "Scope", send: "Send", cookie_keys: List[str]
    ) -> Callable[["Message"], Awaitable[None]]:
        """
        Creates a wrapper for the ASGI send function, which handles setting the cookies on the outgoing response.
        Args:
            scope: The ASGI connection scope.
            send: The ASGI send function.
            cookie_keys: Session cookie keys that are sent in the current request. It is required to expire all session
                cookies from the current request and are replaced with new cookies in the upcoming response.

        Returns:
            None.
        """

        async def wrapped_send(message: "Message") -> None:
            """A wrapper around the send function, declared in local scope to
            use closure values.

            Args:
                message: An ASGI message.

            Returns:
                None
            """
            if message["type"] != "http.response.start":
                await send(message)
                return

            headers = MutableHeaders(scope=message)
            scope_session = scope.get("session")

            if scope_session:
                data = self.dump_data(scope_session, scope=scope)
                cookie_params = self.config.dict(exclude_none=True, exclude={"secret", "key"})
                for i, datum in enumerate(data, start=0):
                    headers.append(
                        "Set-Cookie",
                        Cookie(value=datum.decode("utf-8"), key=f"{self.config.key}-{i}", **cookie_params).to_header(
                            header=""
                        ),
                    )
                # Cookies with the same key overwrite the earlier cookie with that key. To expire earlier session
                # cookies, first check how many session cookies will not be overwritten in this upcoming response.
                # If leftover cookies are greater than or equal to 1, that means older session cookies have to be
                # expired and their names are in cookie_keys.
                cookies_to_clear = cookie_keys[len(data) :] if len(cookie_keys) - len(data) > 0 else []
            else:
                cookies_to_clear = cookie_keys

            for cookie_key in cookies_to_clear:
                cookie_params = self.config.dict(exclude_none=True, exclude={"secret", "max_age", "key"})
                headers.append(
                    "Set-Cookie",
                    Cookie(value="null", key=cookie_key, expires=0, **cookie_params).to_header(header=""),
                )

            await send(message)

        return wrapped_send

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """
        Args:
            scope: The ASGI connection scope.
            receive: The ASGI receive function.
            send: The ASGI send function.

        Returns:
            None
        """
        if scope["type"] in self.config.scopes:
            scope.setdefault("session", {})
            connection = ASGIConnection[Any, Any, Any](scope)
            cookie_keys = sorted(key for key in connection.cookies if self.config.key in key)
            if cookie_keys:
                data = [connection.cookies[key].encode("utf-8") for key in cookie_keys]
                # If these exceptions occur, the session must remain empty so do nothing.
                with contextlib.suppress(InvalidTag, binascii.Error):
                    scope["session"] = self.load_data(data)
            await self.app(scope, receive, self.create_send_wrapper(scope, send, cookie_keys))
        else:
            await self.app(scope, receive, send)



from dataclasses import MISSING
from dataclasses import Field as DataclassField
from dataclasses import dataclass, is_dataclass
from inspect import getmro
from typing import Any, Dict, Generic, Optional, Tuple, Type, TypeVar, get_type_hints

from pydantic import BaseModel, create_model

from starlite.exceptions import ImproperlyConfiguredException
from starlite.utils.predicates import is_class_and_subclass

try:
    # python 3.9 changed these variable
    from typing import _UnionGenericAlias as GenericAlias  # type: ignore
except ImportError:  # pragma: no cover
    from typing import _GenericAlias as GenericAlias  # type: ignore

T = TypeVar("T")


class Partial(Generic[T]):
    """Partial is a special typing helper that takes a generic T, which must be
    a dataclass or pydantic model class,

    and returns to static type checkers a version of this T in which all fields - and nested fields - are optional.
    """

    _models: Dict[Type[T], Type[T]] = {}

    def __class_getitem__(cls, item: Type[T]) -> Type[T]:
        """Takes a pydantic model class or a dataclass and returns an all
        optional version of that class.

        Args:
            item: A pydantic model or dataclass class.

        Returns:
            A pydantic model or dataclass.
        """

        if item not in cls._models:
            if is_class_and_subclass(item, BaseModel):
                cls._create_partial_pydantic_model(item=item)
            elif is_dataclass(item):
                cls._create_partial_dataclass(item=item)
            else:
                raise ImproperlyConfiguredException(
                    "The type argument T passed to Partial[T] must be a dataclass or pydantic model class"
                )

        return cls._models[item]  # pyright: ignore

    @classmethod
    def _create_partial_pydantic_model(cls, item: Type[BaseModel]) -> None:
        """Receives a pydantic model class and returns an all optional subclass
        of it.

        Args:
            item: A pydantic model class.

        Returns:
            A pydantic model class.
        """
        field_definitions: Dict[str, Tuple[Any, None]] = {}
        for field_name, field_type in get_type_hints(item).items():
            if not isinstance(field_type, GenericAlias) or type(None) not in field_type.__args__:
                field_definitions[field_name] = (Optional[field_type], None)
            else:
                field_definitions[field_name] = (field_type, None)

        cls._models[item] = create_model(f"Partial{item.__name__}", __base__=item, **field_definitions)  # type: ignore

    @classmethod
    def _create_partial_dataclass(cls, item: Type[T]) -> None:
        """Receives a dataclass class and returns an all optional subclass of
        it.

        Args:
            item: A dataclass class.

        Returns:
            A dataclass class.
        """
        fields: Dict[str, DataclassField] = cls._create_optional_field_map(item)
        partial_type: Type[T] = dataclass(  # pyright: ignore
            type(f"Partial{item.__name__}", (item,), {"__dataclass_fields__": fields})
        )
        annotated_ancestors = [a for a in getmro(partial_type) if hasattr(a, "__annotations__")]
        for ancestor in annotated_ancestors:
            for field_name, annotation in ancestor.__annotations__.items():
                if not isinstance(annotation, GenericAlias) or type(None) not in annotation.__args__:
                    partial_type.__annotations__[field_name] = Optional[annotation]
                else:
                    partial_type.__annotations__[field_name] = annotation

        cls._models[item] = partial_type

    @staticmethod
    def _create_optional_field_map(item: Type[T]) -> Dict[str, DataclassField]:
        """Creates a map of field name to optional dataclass Fields for a given
        dataclass.

        Args:
            item: A dataclass class.

        Returns:
            A map of field name to optional dataclass fields.
        """
        fields: Dict[str, DataclassField] = {}
        for field_name, dataclass_field in item.__dataclass_fields__.items():  # type: ignore[attr-defined]
            if not isinstance(dataclass_field.type, GenericAlias) or type(None) not in dataclass_field.type.__args__:
                dataclass_field.type = Optional[dataclass_field.type]
            if dataclass_field.default_factory is MISSING:
                dataclass_field.default = None if dataclass_field.default is MISSING else dataclass_field.default
            fields[field_name] = dataclass_field
        return fields



from collections import defaultdict
from datetime import date, datetime, time, timedelta
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Type, Union, cast

from starlette.middleware import Middleware as StarletteMiddleware
from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.trustedhost import TrustedHostMiddleware
from typing_extensions import TypedDict

from starlite.asgi import (
    PathParameterTypePathDesignator,
    PathParamNode,
    RouteMapNode,
    StarliteASGIRouter,
)
from starlite.config import AppConfig, CacheConfig, OpenAPIConfig
from starlite.config.logging import get_logger_placeholder
from starlite.connection import Request, WebSocket
from starlite.datastructures.state import State
from starlite.exceptions import ImproperlyConfiguredException, ValidationException
from starlite.handlers.asgi import asgi
from starlite.handlers.http import HTTPRouteHandler
from starlite.middleware.compression.base import CompressionMiddleware
from starlite.middleware.csrf import CSRFMiddleware
from starlite.middleware.exceptions import ExceptionHandlerMiddleware
from starlite.router import Router
from starlite.routes import ASGIRoute, BaseRoute, HTTPRoute, WebSocketRoute
from starlite.signature import SignatureModelFactory
from starlite.utils import as_async_callable_list, join_paths, unique

if TYPE_CHECKING:
    from pydantic_openapi_schema.v3_1_0 import SecurityRequirement
    from pydantic_openapi_schema.v3_1_0.open_api import OpenAPI

    from starlite.asgi import ComponentsSet, PathParamPlaceholderType
    from starlite.config import (
        BaseLoggingConfig,
        CompressionConfig,
        CORSConfig,
        CSRFConfig,
        StaticFilesConfig,
        TemplateConfig,
    )
    from starlite.datastructures.provide import Provide
    from starlite.handlers.base import BaseRouteHandler
    from starlite.plugins.base import PluginProtocol
    from starlite.routes.base import PathParameterDefinition
    from starlite.types import (
        AfterExceptionHookHandler,
        AfterRequestHookHandler,
        AfterResponseHookHandler,
        AnyCallable,
        ASGIApp,
        BeforeMessageSendHookHandler,
        BeforeRequestHookHandler,
        ControllerRouterHandler,
        ExceptionHandlersMap,
        Guard,
        LifeSpanHandler,
        LifeSpanHookHandler,
        LifeSpanReceive,
        LifeSpanScope,
        LifeSpanSend,
        Logger,
        Message,
        Middleware,
        OnAppInitHandler,
        ParametersMap,
        Receive,
        ResponseCookies,
        ResponseHeadersMap,
        ResponseType,
        RouteHandlerType,
        Scope,
        Send,
        SingleOrList,
    )
    from starlite.types.callable_types import GetLogger

DEFAULT_OPENAPI_CONFIG = OpenAPIConfig(title="Starlite API", version="1.0.0")
"""
    The default OpenAPI config used if not configuration is explicitly passed
    to the [Starlite][starlite.app.Starlite] instance constructor.
"""
DEFAULT_CACHE_CONFIG = CacheConfig()
"""
    The default cache config used if not configuration is explicitly passed
    to the [Starlite][starlite.app.Starlite] instance constructor.
"""


class HandlerIndex(TypedDict):
    """This class is used to map route handler names to a mapping of paths +
    route handler.

    It's returned from the 'get_handler_index_by_name' utility method.
    """

    paths: List[str]
    """Full route paths to the route handler."""
    handler: "RouteHandlerType"
    """Route handler instance."""
    identifier: str
    """Unique identifier of the handler. Either equal to the 'name' attribute or the __str__ value of the handler."""


class HandlerNode(TypedDict):
    """This class encapsulates a route handler node."""

    asgi_app: "ASGIApp"
    """ASGI App stack"""
    handler: "RouteHandlerType"
    """Route handler instance."""


class Starlite(Router):
    __slots__ = (
        "_init",
        "_registered_routes",
        "_route_handler_index",
        "_route_mapping",
        "_static_paths",
        "after_exception",
        "after_shutdown",
        "after_startup",
        "allowed_hosts",
        "asgi_handler",
        "asgi_router",
        "before_send",
        "before_shutdown",
        "before_startup",
        "cache",
        "compression_config",
        "cors_config",
        "csrf_config",
        "debug",
        "get_logger",
        "logger",
        "logging_config",
        "on_shutdown",
        "on_startup",
        "openapi_config",
        "openapi_schema",
        "plain_routes",
        "plugins",
        "request_class",
        "route_map",
        "state",
        "static_files_config",
        "template_engine",
        "websocket_class",
    )

    def __init__(
        self,
        route_handlers: List["ControllerRouterHandler"],
        *,
        after_exception: Optional["SingleOrList[AfterExceptionHookHandler]"] = None,
        after_request: Optional["AfterRequestHookHandler"] = None,
        after_response: Optional["AfterResponseHookHandler"] = None,
        after_shutdown: Optional["SingleOrList[LifeSpanHookHandler]"] = None,
        after_startup: Optional["SingleOrList[LifeSpanHookHandler]"] = None,
        allowed_hosts: Optional[List[str]] = None,
        before_request: Optional["BeforeRequestHookHandler"] = None,
        before_send: Optional["SingleOrList[BeforeMessageSendHookHandler]"] = None,
        before_shutdown: Optional["SingleOrList[LifeSpanHookHandler]"] = None,
        before_startup: Optional["SingleOrList[LifeSpanHookHandler]"] = None,
        cache_config: CacheConfig = DEFAULT_CACHE_CONFIG,
        compression_config: Optional["CompressionConfig"] = None,
        cors_config: Optional["CORSConfig"] = None,
        csrf_config: Optional["CSRFConfig"] = None,
        debug: bool = False,
        dependencies: Optional[Dict[str, "Provide"]] = None,
        exception_handlers: Optional["ExceptionHandlersMap"] = None,
        guards: Optional[List["Guard"]] = None,
        logging_config: Optional["BaseLoggingConfig"] = None,
        middleware: Optional[List["Middleware"]] = None,
        on_app_init: Optional[List["OnAppInitHandler"]] = None,
        on_shutdown: Optional[List["LifeSpanHandler"]] = None,
        on_startup: Optional[List["LifeSpanHandler"]] = None,
        openapi_config: Optional[OpenAPIConfig] = DEFAULT_OPENAPI_CONFIG,
        parameters: Optional["ParametersMap"] = None,
        plugins: Optional[List["PluginProtocol"]] = None,
        request_class: Optional[Type["Request"]] = None,
        response_class: Optional["ResponseType"] = None,
        response_cookies: Optional["ResponseCookies"] = None,
        response_headers: Optional["ResponseHeadersMap"] = None,
        security: Optional[List["SecurityRequirement"]] = None,
        static_files_config: Optional[Union["StaticFilesConfig", List["StaticFilesConfig"]]] = None,
        tags: Optional[List[str]] = None,
        template_config: Optional["TemplateConfig"] = None,
        websocket_class: Optional[Type["WebSocket"]] = None,
    ) -> None:
        """The Starlite application.

        `Starlite` is the root level of the app - it has the base path of "/" and all root level
        Controllers, Routers and Route Handlers should be registered on it.

        It inherits from the [Router][starlite.router.Router] class.

        Args:
            after_exception: An application level [exception hook handler][starlite.types.AfterExceptionHookHandler]
                or list thereof.This hook is called after an exception occurs. In difference to exception handlers,
                it is not meant to return a response - only to process the exception (e.g. log it, send it to Sentry etc.).
            after_request: A sync or async function executed after the route handler function returned and the response
                object has been resolved. Receives the response object which may be either an instance of
                [Response][starlite.response.Response] or `starlette.Response`.
            after_response: A sync or async function called after the response has been awaited. It receives the
                [Request][starlite.connection.Request] object and should not return any values.
            after_shutdown: An application level [life-span hook handler][starlite.types.LifeSpanHookHandler] or
                list thereof. This hook is called during the ASGI shutdown, after all callables in the 'on_shutdown'
                list have been called.
            after_startup: An application level [life-span hook handler][starlite.types.LifeSpanHookHandler] or
                list thereof. This hook is called during the ASGI startup, after all callables in the 'on_startup'
                list have been called.
            allowed_hosts: A list of allowed hosts - enables the builtin allowed hosts middleware.
            before_request: A sync or async function called immediately before calling the route handler.
                Receives the [Request][starlite.connection.Request] instance and any non-`None` return value is
                used for the response, bypassing the route handler.
            before_send: An application level [before send hook handler][starlite.types.BeforeMessageSendHookHandler] or
                list thereof. This hook is called when the ASGI send function is called.
            before_shutdown: An application level [life-span hook handler][starlite.types.LifeSpanHookHandler] or
                list thereof. This hook is called during the ASGI shutdown, before any callables in the 'on_shutdown'
                list have been called.
            before_startup: An application level [life-span hook handler][starlite.types.LifeSpanHookHandler] or
                list thereof. This hook is called during the ASGI startup, before any callables in the 'on_startup'
                list have been called.
            cache_config: Configures caching behavior of the application.
            compression_config: Configures compression behaviour of the application, this enabled a builtin or user
                defined Compression middleware.
            cors_config: If set this enables the builtin CORS middleware.
            csrf_config: If set this enables the builtin CSRF middleware.
            debug: If `True`, app errors rendered as HTML with a stack trace.
            dependencies: A string keyed dictionary of dependency [Provider][starlite.datastructures.Provide] instances.
            exception_handlers: A dictionary that maps handler functions to status codes and/or exception types.
            guards: A list of [Guard][starlite.types.Guard] callables.
            logging_config: A subclass of [BaseLoggingConfig][starlite.config.logging.BaseLoggingConfig].
            middleware: A list of [Middleware][starlite.types.Middleware].
            on_app_init: A sequence of [OnAppInitHandler][starlite.types.OnAppInitHandler] instances. Handlers receive
                an instance of [AppConfig][starlite.config.app.AppConfig] that will have been initially populated with
                the parameters passed to [Starlite][starlite.app.Starlite], and must return an instance of same. If more
                than one handler is registered they are called in the order they are provided.
            on_shutdown: A list of [LifeSpanHandler][starlite.types.LifeSpanHandler] called during
                application shutdown.
            on_startup: A list of [LifeSpanHandler][starlite.types.LifeSpanHandler] called during
                application startup.
            openapi_config: Defaults to [DEFAULT_OPENAPI_CONFIG][starlite.app.DEFAULT_OPENAPI_CONFIG]
            parameters: A mapping of [Parameter][starlite.params.Parameter] definitions available to all
                application paths.
            plugins: List of plugins.
            request_class: An optional subclass of [Request][starlite.connection.request.Request] to use for
                http connections.
            response_class: A custom subclass of [starlite.response.Response] to be used as the app's default response.
            response_cookies: A list of [Cookie](starlite.datastructures.Cookie] instances.
            response_headers: A string keyed dictionary mapping [ResponseHeader][starlite.datastructures.ResponseHeader]
                instances.
            route_handlers: A required list of route handlers, which can include instances of
                [Router][starlite.router.Router], subclasses of [Controller][starlite.controller.Controller] or
                any function decorated by the route handler decorators.
            security: A list of dictionaries that will be added to the schema of all route handlers in the application.
                See [SecurityRequirement][pydantic_openapi_schema.v3_1_0.security_requirement.SecurityRequirement] for details.
            static_files_config: An instance or list of [StaticFilesConfig][starlite.config.StaticFilesConfig]
            tags: A list of string tags that will be appended to the schema of all route handlers under the application.
            template_config: An instance of [TemplateConfig][starlite.config.TemplateConfig]
            websocket_class: An optional subclass of [WebSocket][starlite.connection.websocket.WebSocket] to use for
                websocket connections.
        """
        self._registered_routes: Set[BaseRoute] = set()
        self._route_mapping: Dict[str, List[BaseRoute]] = defaultdict(list)
        self._route_handler_index: Dict[str, "RouteHandlerType"] = {}
        self._static_paths: Set[str] = set()
        self.openapi_schema: Optional["OpenAPI"] = None
        self.get_logger: "GetLogger" = get_logger_placeholder
        self.logger: Optional["Logger"] = None
        self.plain_routes: Set[str] = set()
        self.route_map: RouteMapNode = {}
        self.routes: List[BaseRoute] = []
        self.state = State()

        # creates app config object from parameters
        config = AppConfig(
            after_exception=after_exception or [],
            after_request=after_request,
            after_response=after_response,
            after_shutdown=after_shutdown or [],
            after_startup=after_startup or [],
            allowed_hosts=allowed_hosts or [],
            before_request=before_request,
            before_send=before_send or [],
            before_shutdown=before_shutdown or [],
            before_startup=before_startup or [],
            cache_config=cache_config,
            compression_config=compression_config,
            cors_config=cors_config,
            csrf_config=csrf_config,
            debug=debug,
            dependencies=dependencies or {},
            exception_handlers=exception_handlers or {},
            guards=guards or [],
            logging_config=logging_config,
            middleware=middleware or [],
            on_shutdown=on_shutdown or [],
            on_startup=on_startup or [],
            openapi_config=openapi_config,
            parameters=parameters or {},
            plugins=plugins or [],
            request_class=request_class,
            response_class=response_class,
            response_cookies=response_cookies or [],
            response_headers=response_headers or {},
            route_handlers=route_handlers,
            security=security or [],
            static_files_config=static_files_config or [],
            tags=tags or [],
            template_config=template_config,
            websocket_class=websocket_class,
        )
        for handler in on_app_init or []:
            config = handler(config)

        self.after_exception = as_async_callable_list(config.after_exception)
        self.after_shutdown = as_async_callable_list(config.after_shutdown)
        self.after_startup = as_async_callable_list(config.after_startup)
        self.allowed_hosts = config.allowed_hosts
        self.before_send = as_async_callable_list(config.before_send)
        self.before_shutdown = as_async_callable_list(config.before_shutdown)
        self.before_startup = as_async_callable_list(config.before_startup)
        self.cache = config.cache_config.to_cache()
        self.compression_config = config.compression_config
        self.cors_config = config.cors_config
        self.csrf_config = config.csrf_config
        self.debug = config.debug
        self.logging_config = config.logging_config
        self.on_shutdown = config.on_shutdown
        self.on_startup = config.on_startup
        self.openapi_config = config.openapi_config
        self.plugins = config.plugins
        self.request_class = config.request_class or Request
        self.static_files_config = config.static_files_config
        self.template_engine = config.template_config.to_engine() if config.template_config else None
        self.websocket_class = config.websocket_class or WebSocket

        super().__init__(
            after_request=config.after_request,
            after_response=config.after_response,
            before_request=config.before_request,
            dependencies=config.dependencies,
            exception_handlers=config.exception_handlers,
            guards=config.guards,
            middleware=config.middleware,
            parameters=config.parameters,
            path="",
            response_class=config.response_class,
            response_cookies=config.response_cookies,
            response_headers=config.response_headers,
            # route handlers are registered below
            route_handlers=[],
            security=config.security,
            tags=config.tags,
        )
        for plugin in self.plugins:
            plugin.on_app_init(app=self)

        for route_handler in config.route_handlers:
            self.register(route_handler)

        if self.logging_config:
            self.get_logger = self.logging_config.configure()
            self.logger = self.get_logger("starlite")

        if self.openapi_config:
            self.openapi_schema = self.openapi_config.create_openapi_schema_model(self)
            self.register(self.openapi_config.openapi_controller)

        for static_config in (
            self.static_files_config if isinstance(self.static_files_config, list) else [self.static_files_config]
        ):
            self._static_paths.add(static_config.path)
            self.register(asgi(path=static_config.path)(static_config.to_static_files_app()))

        self.asgi_router = StarliteASGIRouter(on_shutdown=self.on_shutdown, on_startup=self.on_startup, app=self)
        self.asgi_handler = self._create_asgi_handler()

    async def __call__(
        self,
        scope: Union["Scope", "LifeSpanScope"],
        receive: Union["Receive", "LifeSpanReceive"],
        send: Union["Send", "LifeSpanSend"],
    ) -> None:
        """The application entry point.

        Lifespan events (startup / shutdown) are sent to the lifespan handler, otherwise the ASGI handler is used

        Args:
            scope: The ASGI connection scope.
            receive: The ASGI receive function.
            send: The ASGI send function.

        Returns:
            None
        """
        scope["app"] = self
        if scope["type"] == "lifespan":
            await self.asgi_router.lifespan(scope, receive, send)  # type: ignore[arg-type]
            return
        scope["state"] = {}
        await self.asgi_handler(scope, receive, self._wrap_send(send=send, scope=scope))  # type: ignore[arg-type]

    def register(self, value: "ControllerRouterHandler") -> None:  # type: ignore[override]
        """Registers a route handler on the app. This method can be used to
        dynamically add endpoints to an application.

        Args:
            value: an instance of [Router][starlite.router.Router], a subclasses of
        [Controller][starlite.controller.Controller] or any function decorated by the route handler decorators.

        Returns:
            None
        """
        routes = super().register(value=value)
        for route in routes:
            route_handlers = self._get_route_handlers(route)
            for route_handler in route_handlers:

                self._create_handler_signature_model(route_handler=route_handler)
                route_handler.resolve_guards()
                route_handler.resolve_middleware()
                if isinstance(route_handler, HTTPRouteHandler):
                    route_handler.resolve_before_request()
                    route_handler.resolve_after_response()
                    route_handler.resolve_response_handler()
            if isinstance(route, HTTPRoute):
                route.create_handler_map()
            elif isinstance(route, WebSocketRoute):
                route.handler_parameter_model = route.create_handler_kwargs_model(route.route_handler)
        self._construct_route_map()

    def get_handler_index_by_name(self, name: str) -> Optional[HandlerIndex]:
        """Receives a route handler name and returns an optional dictionary
        containing the route handler instance and list of paths sorted
        lexically.

        Examples:
            ```python
            from starlite import Starlite, get


            @get("/", name="my-handler")
            def handler() -> None:
                pass


            app = Starlite(route_handlers=[handler])

            handler_index = app.get_handler_index_by_name("my-handler")

            # { "paths": ["/"], "handler" ... }
            ```
        Args:
            name: A route handler unique name.

        Returns:
            A [HandlerIndex][starlite.app.HandlerIndex] instance or None.
        """
        handler = self._route_handler_index.get(name)
        if not handler:
            return None

        identifier = handler.name or str(handler)
        routes = self._route_mapping[identifier]
        paths = sorted(unique([route.path for route in routes]))

        return HandlerIndex(handler=handler, paths=paths, identifier=identifier)

    def route_reverse(self, name: str, **path_parameters: Any) -> Optional[str]:
        """Receives a route handler name, path parameter values and returns an
        optional url path to the handler with filled path parameters.

        Examples:
            ```python
            from starlite import Starlite, get


            @get("/group/{group_id:int}/user/{user_id:int}", name="get_membership_details")
            def get_membership_details(group_id: int, user_id: int) -> None:
                pass


            app = Starlite(route_handlers=[get_membership_details])

            path = app.route_reverse("get_memebership_details", user_id=100, group_id=10)

            # /group/10/user/100
            ```
        Args:
            name: A route handler unique name.
            **path_parameters: Actual values for path parameters in the route.

        Raises:
            ValidationException: If path parameters are missing in **path_parameters or have wrong type.

        Returns:
            A fully formatted url path or None.
        """
        handler_index = self.get_handler_index_by_name(name)
        if handler_index is None:
            return None

        allow_str_instead = {datetime, date, time, timedelta, float, Path}
        output: List[str] = []

        routes = sorted(
            self._route_mapping[handler_index["identifier"]], key=lambda r: len(r.path_parameters), reverse=True
        )
        passed_parameters = set(path_parameters.keys())

        selected_route = routes[-1]
        for route in routes:
            if passed_parameters.issuperset({param["name"] for param in route.path_parameters}):
                selected_route = route
                break

        for component in selected_route.path_components:
            if isinstance(component, dict):
                val = path_parameters.get(component["name"])
                if not (
                    isinstance(val, component["type"])
                    or (component["type"] in allow_str_instead and isinstance(val, str))
                ):
                    raise ValidationException(
                        f"Received type for path parameter {component['name']} doesn't match declared type {component['type']}"
                    )
                output.append(str(val))
            else:
                output.append(component)

        return join_paths(output)

    @property
    def route_handler_method_view(self) -> Dict[str, List[str]]:
        """
        Returns:
            A dictionary mapping route handlers to paths.
        """
        route_map: Dict[str, List[str]] = {}
        for handler, routes in self._route_mapping.items():
            route_map[handler] = [route.path for route in routes]

        return route_map

    def _create_asgi_handler(self) -> "ASGIApp":
        """Creates an ASGIApp that wraps the ASGI router inside an exception
        handler.

        If CORS or TrustedHost configs are provided to the constructor,
        they will wrap the router as well.
        """
        asgi_handler: "ASGIApp" = self.asgi_router
        if self.compression_config:
            asgi_handler = CompressionMiddleware(app=asgi_handler, config=self.compression_config)
        if self.allowed_hosts:
            asgi_handler = TrustedHostMiddleware(app=asgi_handler, allowed_hosts=self.allowed_hosts)  # type: ignore
        if self.cors_config:
            asgi_handler = CORSMiddleware(app=asgi_handler, **self.cors_config.dict())  # type: ignore
        if self.csrf_config:
            asgi_handler = CSRFMiddleware(app=asgi_handler, config=self.csrf_config)
        return self._wrap_in_exception_handler(asgi_handler, exception_handlers=self.exception_handlers or {})

    def _wrap_in_exception_handler(self, app: "ASGIApp", exception_handlers: "ExceptionHandlersMap") -> "ASGIApp":
        """Wraps the given ASGIApp in an instance of
        ExceptionHandlerMiddleware."""
        return ExceptionHandlerMiddleware(app=app, exception_handlers=exception_handlers, debug=self.debug)

    def _add_node_to_route_map(self, route: BaseRoute) -> RouteMapNode:
        """Adds a new route path (e.g. '/foo/bar/{param:int}') into the
        route_map tree.

        Inserts non-parameter paths ('plain routes') off the tree's root
        node. For paths containing parameters, splits the path on '/'
        and nests each path segment under the previous segment's node
        (see prefix tree / trie).
        """
        current_node = self.route_map
        path = route.path

        if route.path_parameters or path in self._static_paths:
            components = cast(
                "List[Union[str, PathParamPlaceholderType, PathParameterDefinition]]", ["/", *route.path_components]
            )
            for component in components:
                components_set = cast("ComponentsSet", current_node["_components"])

                if isinstance(component, dict):
                    # The rest of the path should be regarded as a parameter value.
                    if component["type"] is Path:
                        components_set.add(PathParameterTypePathDesignator)
                    # Represent path parameters using a special value
                    component = PathParamNode

                components_set.add(component)

                if component not in current_node:
                    current_node[component] = {"_components": set()}
                current_node = cast("RouteMapNode", current_node[component])
                if "_static_path" in current_node:
                    raise ImproperlyConfiguredException("Cannot have configured routes below a static path")
        else:
            if path not in self.route_map:
                self.route_map[path] = {"_components": set()}
            self.plain_routes.add(path)
            current_node = self.route_map[path]
        self._configure_route_map_node(route, current_node)
        return current_node

    @staticmethod
    def _get_route_handlers(route: BaseRoute) -> List["RouteHandlerType"]:
        """Retrieve handler(s) as a list for given route."""
        route_handlers: List["RouteHandlerType"] = []
        if isinstance(route, (WebSocketRoute, ASGIRoute)):
            route_handlers.append(route.route_handler)
        else:
            route_handlers.extend(cast("HTTPRoute", route).route_handlers)

        return route_handlers

    def _store_handler_to_route_mapping(self, route: BaseRoute) -> None:
        """Stores the mapping of route handlers to routes and to route handler
        names.

        Args:
            route: A Route instance.

        Returns:
            None
        """
        route_handlers = self._get_route_handlers(route)

        for handler in route_handlers:
            if handler.name in self._route_handler_index and str(self._route_handler_index[handler.name]) != str(
                handler
            ):
                raise ImproperlyConfiguredException(
                    f"route handler names must be unique - {handler.name} is not unique."
                )
            identifier = handler.name or str(handler)
            self._route_mapping[identifier].append(route)
            self._route_handler_index[identifier] = handler

    def _configure_route_map_node(self, route: BaseRoute, node: RouteMapNode) -> None:
        """Set required attributes and route handlers on route_map tree
        node."""
        if "_path_parameters" not in node:
            node["_path_parameters"] = route.path_parameters
        if "_asgi_handlers" not in node:
            node["_asgi_handlers"] = {}
        if "_is_asgi" not in node:
            node["_is_asgi"] = False
        if route.path in self._static_paths:
            if node["_components"]:
                raise ImproperlyConfiguredException("Cannot have configured routes below a static path")
            node["_static_path"] = route.path
            node["_is_asgi"] = True
        asgi_handlers = cast("Dict[str, HandlerNode]", node["_asgi_handlers"])
        if isinstance(route, HTTPRoute):
            for method, handler_mapping in route.route_handler_map.items():
                handler, _ = handler_mapping
                asgi_handlers[method] = HandlerNode(
                    asgi_app=self._build_route_middleware_stack(route, handler),
                    handler=handler,
                )
        elif isinstance(route, WebSocketRoute):
            asgi_handlers["websocket"] = HandlerNode(
                asgi_app=self._build_route_middleware_stack(route, route.route_handler),
                handler=route.route_handler,
            )
        elif isinstance(route, ASGIRoute):
            asgi_handlers["asgi"] = HandlerNode(
                asgi_app=self._build_route_middleware_stack(route, route.route_handler),
                handler=route.route_handler,
            )
            node["_is_asgi"] = True

    def _construct_route_map(self) -> None:
        """Create a map of the app's routes.

        This map is used in the asgi router to route requests.
        """
        if "_components" not in self.route_map:
            self.route_map["_components"] = set()
        new_routes = [route for route in self.routes if route not in self._registered_routes]
        for route in new_routes:
            node = self._add_node_to_route_map(route)
            if node["_path_parameters"] != route.path_parameters:
                raise ImproperlyConfiguredException("Should not use routes with conflicting path parameters")
            self._store_handler_to_route_mapping(route)
            self._registered_routes.add(route)

    def _build_route_middleware_stack(
        self,
        route: Union[HTTPRoute, WebSocketRoute, ASGIRoute],
        route_handler: "RouteHandlerType",
    ) -> "ASGIApp":
        """Constructs a middleware stack that serves as the point of entry for
        each route."""

        # we wrap the route.handle method in the ExceptionHandlerMiddleware
        asgi_handler = self._wrap_in_exception_handler(
            app=route.handle, exception_handlers=route_handler.resolve_exception_handlers()  # type: ignore[arg-type]
        )

        for middleware in route_handler.resolve_middleware():
            if isinstance(middleware, StarletteMiddleware):
                handler, kwargs = middleware
                asgi_handler = handler(app=asgi_handler, **kwargs)
            else:
                asgi_handler = middleware(app=asgi_handler)  # type: ignore

        # we wrap the entire stack again in ExceptionHandlerMiddleware
        return self._wrap_in_exception_handler(
            app=asgi_handler, exception_handlers=route_handler.resolve_exception_handlers()  # pyright: ignore
        )

    def _create_handler_signature_model(self, route_handler: "BaseRouteHandler") -> None:
        """Creates function signature models for all route handler functions
        and provider dependencies."""
        if not route_handler.signature_model:
            route_handler.signature_model = SignatureModelFactory(
                fn=cast("AnyCallable", route_handler.fn),
                plugins=self.plugins,
                dependency_names=route_handler.dependency_name_set,
            ).create_signature_model()
        for provider in list(route_handler.resolve_dependencies().values()):
            if not provider.signature_model:
                provider.signature_model = SignatureModelFactory(
                    fn=provider.dependency,
                    plugins=self.plugins,
                    dependency_names=route_handler.dependency_name_set,
                ).create_signature_model()

    def _wrap_send(self, send: "Send", scope: "Scope") -> "Send":
        """Wraps the ASGI send and handles any 'before send' hooks.

        Args:
            send: The ASGI send function.

        Returns:
            An ASGI send function.
        """
        if self.before_send:

            async def wrapped_send(message: "Message") -> None:
                for hook in self.before_send:
                    if hook.num_expected_args > 2:
                        await hook(message, self.state, scope)
                    else:
                        await hook(message, self.state)
                await send(message)

            return wrapped_send
        return send



from typing import Any

from typing_extensions import Protocol


class Logger(Protocol):  # pragma: no cover
    def debug(self, event: str, **kwargs: Any) -> Any:
        """Outputs a log message at 'DEBUG' level.

        Args:
             event: Log message.
             **kwargs: Any kwargs.
        """

    def info(self, event: str, **kwargs: Any) -> Any:
        """Outputs a log message at 'INFO' level.

        Args:
             event: Log message.
             **kwargs: Any kwargs.
        """

    def warning(self, event: str, **kwargs: Any) -> Any:
        """Outputs a log message at 'WARN' level.

        Args:
             event: Log message.
             **kwargs: Any kwargs.
        """

    def warn(self, event: str, **kwargs: Any) -> Any:
        """Outputs a log message at 'WARN' level.

        Args:
             event: Log message.
             **kwargs: Any kwargs.
        """

    def error(self, event: str, **kwargs: Any) -> Any:
        """Outputs a log message at 'ERROR' level.

        Args:
             event: Log message.
             **kwargs: Any kwargs.
        """

    def fatal(self, event: str, **kwargs: Any) -> Any:
        """Outputs a log message at 'CRITICAL' level.

        Args:
             event: Log message.
             **kwargs: Any kwargs.
        """

    def exception(self, event: str, **kwargs: Any) -> Any:
        """Logs a message with level 'ERROR' on this logger. The arguments are
        interpreted as for debug(). Exception info is added to the logging
        message.

        Args:
             event: Log message.
             **kwargs: Any kwargs.
        """

    def critical(self, event: str, **kwargs: Any) -> Any:
        """Outputs a log message at 'CRITICAL' level.

        Args:
             event: Log message.
             **kwargs: Any kwargs.
        """



from contextlib import suppress
from functools import reduce
from typing import TYPE_CHECKING, Any, Dict, List, Tuple
from urllib.parse import parse_qsl

from orjson import JSONDecodeError, loads
from pydantic.fields import SHAPE_LIST, SHAPE_SINGLETON
from starlite_multipart.datastructures import UploadFile as MultipartUploadFile

from starlite.datastructures.upload_file import UploadFile
from starlite.enums import RequestEncodingType

if TYPE_CHECKING:

    from pydantic.fields import ModelField

    from starlite.datastructures.form_multi_dict import FormMultiDict

_true_values = {"True", "true"}
_false_values = {"False", "false"}


def _query_param_reducer(acc: Dict[str, List[Any]], cur: Tuple[str, str]) -> Dict[str, List[str]]:
    """
    Reducer function - acc is a dictionary, cur is a tuple of key + value

    We use reduce because python implements reduce in C, which makes it faster than a regular for loop in most cases.
    """
    key, value = cur

    if value in _true_values:
        value = True  # type: ignore
    elif value in _false_values:
        value = False  # type: ignore

    if key in acc:
        acc[key].append(value)
    else:
        acc[key] = [value]
    return acc


def parse_query_params(query_string: bytes) -> Dict[str, List[str]]:
    """Parses and normalize a given connection's query parameters into a
    regular dictionary.

    Args:
        query_string: A byte-string containing a query

    Returns:
        A string keyed dictionary of values.
    """

    return reduce(_query_param_reducer, parse_qsl(query_string.decode("utf-8"), keep_blank_values=True), {})


def parse_form_data(media_type: "RequestEncodingType", form_data: "FormMultiDict", field: "ModelField") -> Any:
    """Transforms the multidict into a regular dict, try to load json on all
    non-file values.

    Supports lists.
    """
    values_dict: Dict[str, Any] = {}
    for key, value in form_data.multi_items():
        if not isinstance(value, MultipartUploadFile):
            with suppress(JSONDecodeError):
                value = loads(value)
        existing_value = values_dict.get(key)
        if isinstance(existing_value, list):
            values_dict[key].append(value)
        elif existing_value:
            values_dict[key] = [existing_value, value]
        else:
            values_dict[key] = value
    if media_type == RequestEncodingType.MULTI_PART:
        if field.shape is SHAPE_LIST:
            return list(values_dict.values())
        if field.shape is SHAPE_SINGLETON and field.type_ in (UploadFile, MultipartUploadFile) and values_dict:
            return list(values_dict.values())[0]
    return values_dict



from typing import Any

from starlette.datastructures import ImmutableMultiDict

from starlite.datastructures.upload_file import UploadFile


class FormMultiDict(ImmutableMultiDict[str, Any]):
    async def close(self) -> None:
        """Closes all files in the multi-dict.

        Returns:
            None
        """
        for _, value in self.multi_items():
            if isinstance(value, UploadFile):
                await value.close()



from typing import Awaitable, List, TypeVar, Union

T = TypeVar("T")

SyncOrAsyncUnion = Union[T, Awaitable[T]]
"""
Types 'T' as a union of T and awaitable T
"""
SingleOrList = Union[T, List[T]]
"""
Types 'T' as a single value or a list T
"""



from starlite.datastructures.background_tasks import BackgroundTask, BackgroundTasks
from starlite.datastructures.cookie import Cookie
from starlite.datastructures.form_multi_dict import FormMultiDict
from starlite.datastructures.provide import Provide
from starlite.datastructures.response_containers import (
    File,
    Redirect,
    ResponseContainer,
    Stream,
    Template,
)
from starlite.datastructures.response_header import ResponseHeader
from starlite.datastructures.state import State
from starlite.datastructures.upload_file import UploadFile

__all__ = (
    "BackgroundTask",
    "BackgroundTasks",
    "Cookie",
    "File",
    "FormMultiDict",
    "Provide",
    "Redirect",
    "ResponseContainer",
    "ResponseHeader",
    "State",
    "Stream",
    "Template",
    "UploadFile",
)
